# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List

from cosmos_rl.rollout.schema import RolloutResult

from cosmos_rl.policy.config import Config as CosmosConfig
from transformers import AutoTokenizer


class RolloutBase(ABC):
    def __init__(self, config: CosmosConfig, tokenizer: AutoTokenizer):
        self.config = config
        self.tokenizer = tokenizer
        self.pad_token_id = self.tokenizer.pad_token_id

    @abstractmethod
    def rollout_generation(self, prompts, *args, **kwargs) -> List[RolloutResult]:
        """Generate sequences"""
        pass

    @abstractmethod
    def init_engine(self, quantization: str, seed: int, load_format: str):
        """Initialize the engine"""
        pass
