# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple
from abc import ABC, abstractmethod

from .schema import OpenAIFunctionCallSchema


class ToolParser(ABC):
    @abstractmethod
    def extract_tool_calls(
        self, text: str
    ) -> Tuple[str, list[OpenAIFunctionCallSchema]]:
        """Extract tool calls from the text.

        Args:
            text (str): The responses text.

        Returns:
            Tuple[str, List[OpenAIFunctionCallSchema]]: Content and extracted tool calls.
        """
        raise NotImplementedError
