# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# patch trtllm according to trtllm version
import tensorrt_llm

trtllm_version = tensorrt_llm.__version__

if trtllm_version == "1.0.0rc6":
    from cosmos_rl.rollout.trtllm_rollout.patches_for_trtllm import trtllm_1_0_0_rc6  # noqa: F401
else:
    raise NotImplementedError(
        f"[Rollout] Unsupported trtllm version: {trtllm_version}."
    )
